from . import fmt
from . import storage
from .decorators import savedata, savefig
from .paths import add_arg, delete_arg, modify_arg

__all__ = ["savedata",
           "savefig",
           'fmt',
           'storage',
           "add_args",
           'delete_arg',
           'modify_arg'
          ]


"""
Helper functions
"""
import re

def merge_nested_dict(d, keys, key_default=None):
    """
    Returns the result of merging several keys of a dictionary.
    
    Attrs:
            - d:             nested dictionary
            - keys:          d keys to be merged. can be a string of the form 'k' or 'k1+k2+...', or other iterables containing the keys.
            - key_default:   if a key does not belong to d => it is searched in d[key_default]
    """
    if isinstance(keys, str):
        key_add = set(re.findall(r'(?:^|(?<=\+))(\w+)', keys))
        key_remove = set(re.findall(r'(?<=-)(\w+)', keys))
    else:
        key_add = set(keys)
        key_remove = set()
    
    if not key_add and key_remove:
        key_add = set(["all"])
        
    d_keys = set(d)
    keys_in_d = key_add.intersection(d_keys)
    keys_in_default = key_add - keys_in_d
    
    if not keys_in_default.issubset(set(d[key_default])):
        raise RuntimeError("Some keys don't belong to d or d[key_default]")
    else:    
        d_merged = {}    
        for k in keys_in_d:
            d_merged.update(d[k])
        d_default = d[key_default]
        d_merged.update({k: d_default[k] for k in keys_in_default})
        for k in key_remove:
            if k in d_merged:
                del d_merged[k]
            else:
                for k_i in d[k]:
                    del d_merged[k_i]
        return d_merged

class NoFigure():
    """
    Fill type for plotly/matplotlib figures, if the package is not available.
    Allows tidypath.savefig to work with matplotlib/plotly when the other one is not present.
    """
    pass


"""
Main uses: save and cache data, save figures, change figure color, timing.
"""
import math
import inspect
import warnings
import os
from pathlib import Path
from functools import wraps
from collections.abc import Iterable
from importlib.util import find_spec
from ._helper import NoFigure
from lzma import LZMAError
if find_spec("plotly") is None:
    plotly_figure = NoFigure
else:
    from plotly.graph_objs._figure import Figure as plotly_figure
if find_spec("matplotlib") is None:
    mpl_figure = NoFigure
else:
    from matplotlib.figure import Figure as mpl_figure
    import matplotlib.pyplot as plt

from . import storage, config
from .paths import datapath, figpath, hash_path
from .inspection import classify_call_attrs, merge_wrapper_signatures
from ._helper import merge_nested_dict


def savedata(keys_or_function=None, include_classes="file",
             ext=config.EXT_DEFAULT_DATA, keys=config.KEYS_DEFAULT_DATA, funcname_in_filename=config.FUNCNAME_IN_FILENAME_DEFAULT_DATA,
             overwrite=False, save=True, load_opts_default_save=True,  #defaults for extra arguments
             max_str_length=255,
             iterable_maxsize=math.inf,
             load_opts={}, **save_opts):
    """
    Decorator for automatically saving output and then loading cached data.
    Default behavior:
        1st function call:   stores the data with extension 'ext'.
        rest:                loads stored data if the key args passed to the function coincide.

    If filename is too long, it is shortened by hashing it.


    NOTE (!) decorated funcs will have extra arguments:
        - save (bool).                 whether to save the output of the function.
        - overwrite (bool).            whether to overwrite the saved version.
        - funcname_in_filename(bool)   whether to include funcname in the filename.
        - keys (dict/str/iterable)   · dict:     specifies filename of the form k1-v1_k2-v2_...kn_vn.
                                                 k_i do not have to be arguments of the function.
                                     · str:       key of a keyword argument. Example: keys = 'x'.
                                                 "kwargs_defaults"  =>  default kwargs that were not modified.
                                                 "kwargs"           =>  kwargs passed during function call.
                                                 "kwargs_full"      =>  kws_defaults + kws.
                                                 "pos_only"         =>  length of *args
                                                 "args"             =>  attrs != **kwargs, *args.
                                                 "all"              if config.KEYS_ADD_POSONLY_TO_ALL  =>  all attrs
                                                                    else                               =>  all attrs except pos_only: kwargs_full + args
                                                 can combine options using "+" or "-". Example: "args+z", "x+y+kwargs", "all-y".
                                     · iterable: containing a combination of the available string keys (above). ["k1", "k2"] == "k1+k2".

    Attrs:
        - function:                function to which the decorator is applied
        - ext:                     storing extension. Selects 'storage' functions save_ext, load_ext.
                                   Supported: 'lzma' (default), 'bz2', 'json', 'csv', 'npz'.
        - include_classes:         include class tree in saving_path.
        - load_opts:               kws for storage.load_ext. default kws are those of 'saving_options', those specified update saving_options dict.
        - save_opts:               kws for storage.save_ext
        - load_opts_default_save:  use save_opts as default for load_opts.
        - max_str_length:          max length of filename. If exceeded, filename is shortened by hashing it.
        - iterable_maxsize:        max size of iterable keys. If exceeded, keys are shortened by counting val numbers.
        - rest:                    default behavior for decorated funcs extra arguments (above).

    Returns: Function decorator
    """
    if isinstance(keys_or_function, Iterable):
        keys = keys_or_function
        func = None
    elif callable(keys_or_function):
        func = keys_or_function
    else:
        func = None

    if load_opts_default_save:
        load_opts = {**save_opts, **load_opts}

    def _savedata(func):
        @wraps(func)
        def wrapper(*args, overwrite=overwrite, keys=keys, save=save, funcname_in_filename=funcname_in_filename, **kwargs):
            key_opts = classify_call_attrs(func, args, kwargs, add_pos_only_to_all=config.KEYS_ADD_POSONLY_TO_ALL)
            save_keys = merge_nested_dict(key_opts, keys, key_default="all")
            saving_path = datapath(keys=save_keys, func=func, ext=ext, include_classes=include_classes, funcname_in_filename=funcname_in_filename, iterable_maxsize=iterable_maxsize)

            filename_too_long = len(saving_path) > max_str_length
            if filename_too_long:
                saving_path = hash_path(saving_path)

            if Path(saving_path).exists() and not overwrite:
                try:
                    result = getattr(storage, f"load_{ext}")(saving_path, **load_opts)
                except EOFError or LZMAError:
                    warnings.warn("Corrupted file. Recomputing and storing ...", RuntimeWarning)
                    result = func(*args, **kwargs)
                    getattr(storage, f"save_{ext}")(result, saving_path, **save_opts)
            else:
                if filename_too_long:
                    warnings.warn("Filename too long. Hashing it ...", RuntimeWarning)
                result = func(*args, **kwargs)
                getattr(storage, f"save_{ext}")(result, saving_path, **save_opts)
            return result

        wrapper.__signature__ = merge_wrapper_signatures(wrapper, ["overwrite", "keys", "save", "funcname_in_filename"])
        wrapper.__out__ = "data"
        return wrapper

    if func is None:
        return _savedata
    else:
        return _savedata(func)


def savefig(keys_or_function=None, include_classes="file",
            ext=config.EXT_DEFAULT_FIG, keys=config.KEYS_DEFAULT_FIG, funcname_in_filename=config.FUNCNAME_IN_FILENAME_DEFAULT_FIG, return_fig=config.RETURN_FIG_DEFAULT,
            max_str_length=255,
            iterable_maxsize=3,
            overwrite=True, save=True,  #defaults for extra arguments
            **save_opts):
    """
    Generates figsaver decorator.
    Figure returned by function is automatically saved. Compatible with matplotlib and plotly.
    If None or NaN is returned, avoid figure saving.

    If filename is too long, it is shortened by hashing it.

    NOTE (!) decorated funcs will have extra arguments:
        - return_fig (bool)            whether to return the figure (output of function).
        - save (bool).                 whether to save the figure.
        - overwrite (bool).            whether to overwrite the saved version.
        - funcname_in_filename(bool)   whether to include funcname in the filename.
        - keys (dict/str/iterable)   · dict:     specifies filename of the form k1-v1_k2-v2_...kn_vn.
                                                 k_i do not have to be arguments of the function.
                                     · str:       key of a keyword argument. Example: keys = 'x'.
                                                 "kwargs_defaults"  =>  default kwargs that were not modified.
                                                 "kwargs"           =>  kwargs passed during function call.
                                                 "kwargs_full"      =>  kws_defaults + kws.
                                                 "pos_only"         =>  length of *args. Also self and cls are counted as pos_only arguments.
                                                 "args"             =>  attrs != **kwargs, *args.
                                                 "all"              if config.KEYS_ADD_POSONLY_TO_ALL  =>  all attrs
                                                                    else                               =>  all attrs except pos_only: kwargs_full + args
                                                 can combine options using "+" or "-". Example: "args+z", "x+y+kwargs", "all-y".
                                     · iterable: containing a combination of the available string keys (above). ["k1", "k2"] == "k1+k2".

    Attrs:
        - function:                function to which the decorator is applied
        - ext:                     storing extension. 'eps' recommended for articles.
                                   Supported: any extension supported by matplotlib/plotly. Example: 'png', 'eps', 'html' (plotly), etc.
        - include_classes:         include class tree in saving_path.
        - save_opts:               kws for saving function.
        - max_str_length:          max length of filename. If exceeded, filename is shortened by hashing it.
        - iterable_maxsize:        max size of iterable keys. If exceeded, keys are shortened by counting val numbers.
        - rest:                    default behavior for decorated funcs extra arguments (above).

    Returns: Function decorator
    """
    if isinstance(keys_or_function, Iterable):
        keys = keys_or_function
        func = None
    elif callable(keys_or_function):
        func = keys_or_function
    else:
        func = None

    mpl_save_defaults = dict(bbox_inches="tight")

    def _savefig(func):
        @wraps(func)
        def wrapper(*args, overwrite=overwrite, keys=keys, save=save, return_fig=return_fig, funcname_in_filename=funcname_in_filename, **kwargs):
            fig = func(*args, **kwargs)
            if isinstance(fig, (mpl_figure, plotly_figure)) or type(fig).__name__ == 'AxesSubplot':
                key_opts = classify_call_attrs(func, args, kwargs, add_pos_only_to_all=config.KEYS_ADD_POSONLY_TO_ALL)
                save_keys = merge_nested_dict(key_opts, keys, key_default="all")
                saving_path = figpath(keys=save_keys, func=func, ext=ext, include_classes=include_classes, funcname_in_filename=funcname_in_filename, iterable_maxsize=iterable_maxsize)

                if len(saving_path) > max_str_length:
                    saving_path = hash_path(saving_path)
                    warnings.warn("Filename too long. Hashing it ...", RuntimeWarning)

                if not Path(saving_path).exists() or overwrite:
                    if isinstance(fig, mpl_figure) or type(fig).__name__ == 'AxesSubplot':
                        fig.savefig(saving_path, format=ext, **{**mpl_save_defaults, **save_opts})
                        plt.close(fig)
                    elif isinstance(fig, plotly_figure):
                        if ext == "html":
                            fig.write_html(saving_path, **save_opts)
                        else:
                            fig.write_image(saving_path, format=ext, **save_opts)
                    else:
                        raise TypeError(f"fig type '{type(fig)}' not valid. Available: 'matplotlib.figure.Figure', 'matplotlib.axes._subplots.AxesSubplot', 'plotly.grap_objs._figure.Figure'.")

                if return_fig:
                    return fig
                else:
                    return
            elif fig is None or math.isnan(fig):
                warnings.warn("Expected output figure (plotly or matplotlib) and received None or NaN.", RuntimeWarning)
            else:
                warnings.warn("Expected output figure (plotly, matplotlib) or a flag for figure error computation (None, NaN), but received {}".format(type(fig)), RuntimeWarning)

        wrapper.__signature__ = merge_wrapper_signatures(wrapper, ["overwrite", "keys", "save", "funcname_in_filename", "return_fig"])
        wrapper.__out__ = "figure"
        return wrapper

    if func is None:
        return _savefig
    else:
        return _savefig(func)



"""Config specs for the decorators 'savedata', 'savefig'."""

import os
import warnings

# YAML needed to use file based Numba config
try:
    import yaml
    _HAVE_YAML = True
except ImportError:
    _HAVE_YAML = False

# this is the name of the user supplied configuration file
_config_fname = '.tidypath_config.yaml'    
    
class _EnvReloader(object):

    def __init__(self):
        self.reset()

    def reset(self):
        self.old_environ = {}
        self.update(force=True)

    def update(self, force=False):
        new_environ = {}

        # first check if there's a .phdu_config.yaml and use values from that
        if os.path.exists(_config_fname) and os.path.isfile(_config_fname):
            if not _HAVE_YAML:
                msg = ("A tidypath config file is found but YAML parsing "
                       "capabilities appear to be missing. "
                       "To use this feature please install `pyyaml`. e.g. "
                       "`conda install pyyaml`.")
                warnings.warn(msg)
            else:
                with open(_config_fname, 'rt') as f:
                    y_conf = yaml.safe_load(f)
                if y_conf is not None:
                    for k, v in y_conf.items():
                        new_environ['TIDYPATH_' + k.upper()] = v

        # clobber file based config with any locally defined env vars
        for name, value in os.environ.items():
            if name.startswith('TIDYPATH_'):
                new_environ[name] = value
        # We update the config variables if at least one PDHU environment
        # variable was modified.  This lets the user modify values
        # directly in the config module without having them when
        # reload_config() is called by the compiler.
        if force or self.old_environ != new_environ:
            self.process_environ(new_environ)
            # Store a copy
            self.old_environ = dict(new_environ)

    def process_environ(self, environ):
        def _readenv(name, ctor, default):
            """
            Attrs:
                    - name:    name of the environment variable
                    - ctor:    constructor for the value of the env var. Takes a value v1 of type t1 and returns a value v2 of type t2.
                    - default: default value for the env var.
            """
            value = environ.get(name)
            if value is None:
                return default() if callable(default) else default
            try:
                return ctor(value)
            except Exception:
                warnings.warn("environ %s defined but failed to parse '%s'" %
                              (name, value), RuntimeWarning)
                return default

        def optional_str(x):
            """
            Examples:
                        ENV_VAR = _readenv(TIDYPATH_ENV_VAR, optional_str, None)
                        ENV_VAR = _readenv(TIDYPATH_ENV_VAR, optional_str,
                                           (val if condition 
                                            else None))
            """
            return str(x) if x is not None else None
        
        def to_bool(x):
            if isinstance(x, str):
                if x.lower() in ["true", "t", "y", "yes"]:
                    return True
                elif x.lower() in ["false", "f", "n", "no"]:
                    return False
                else:
                    raise ValueError(f"x:str must be one of ['true', 'false', yes', no'] and the corresponding 1 character equivalents. Received {x}")
            elif isinstance(x, int):
                if x > 1:
                    raise ValueError(f"x:int must be 0 or 1. Received {x}")
                return x == 1
            else:
                raise TypeError("x must be of type 'str' or 'int'. Received {}".format(type(x)))
        
        KEYS_DEFAULT = _readenv("TIDYPATH_KEYS_DEFAULT", str, "kwargs_full")
        KEYS_DEFAULT_DATA = _readenv("TIDYPATH_KEYS_DEFAULT_DATA", str, KEYS_DEFAULT)
        KEYS_DEFAULT_FIG = _readenv("TIDYPATH_KEYS_DEFAULT_FIG", str, KEYS_DEFAULT)
        KEYS_ADD_POSONLY_TO_ALL = _readenv("TIDYPATH_KEYS_ADD_POSONLY_TO_ALL", to_bool, False)
        
        EXT_DEFAULT_DATA = _readenv("TIDYPATH_EXT_DEFAULT_DATA", str, "lzma")
        EXT_DEFAULT_FIG = _readenv("TIDYPATH_EXT_DEFAULT_FIG", str, "png")
        
        FUNCNAME_IN_FILENAME_DEFAULT_DATA = _readenv("TIDYPATH_FUNCNAME_IN_FILENAME_DEFAULT_DATA", to_bool, False)
        FUNCNAME_IN_FILENAME_DEFAULT_FIG = _readenv("TIDYPATH_FUNCNAME_IN_FILENAME_DEFAULT_FIG", to_bool, True)
        
        RETURN_FIG_DEFAULT = _readenv("TIDYPATH_RETURN_FIG_DEFAULT", to_bool, False)
        
        # Inject the configuration values into the module globals
        for name, value in locals().copy().items():
            if name.isupper():
                globals()[name] = value

_env_reloader = _EnvReloader()


def reload_config():
    """
    Reload the configuration from environment variables, if necessary.
    """
    _env_reloader.update()


"""
Analyze variables passed to functions, parent class, merge wrapper and wrapped signatures.
"""
import inspect
import functools

def classify_call_attrs(func, args, kwargs, add_pos_only_to_all=False):
    """
    Classify function args and kwargs passed during function call
    
    NOTE:  Probably could be done more efficiently relying more on inspect.getargspec(func) or inspect.signature(func), but realized later.
    NOTE2: For position-only arguments f(*args), 
           usually referred in the docs as f(pos1, pos2, /, pos_or_kwd,...)  ('/' indicates end of position-only-args),
           only provides the number of *args.
    
    Example: dict()
    
    Returns: dict containing
        "kwargs_defaults"  =>  default kwargs that were not modified.
        "kwargs"           =>  kwargs passed during function call.
        "kwargs_full"      =>  kws_defaults + kws.
        "pos_only"         =>  length of *args (position-only arguments)
        "args"             =>  position-or-keyword arguments (are called static in the code).
        "all"              =>  add_pos_only_to_all == False => all attrs except pos_only: kwargs_full + args
                               else                         => all attrs: kwargs_full + args + pos_only
    """
    code = func.__code__
    num_kwonly = code.co_kwonlyargcount
    if num_kwonly > 0: # func has keyword-only args: f(*args, k1=v1, ..., kn=vn)
        kwonly_defaults = func.__kwdefaults__
    else:
        kwonly_defaults = {}

    num_defaults = 0 if func.__defaults__ is None else len(func.__defaults__) # f(k1=v1,...)
    num_static = code.co_argcount - num_defaults                              # f(k1,k2)
    if num_defaults > 0:
        args_default = code.co_varnames[:code.co_argcount][-num_defaults:]
        kwargs_defaults = dict(zip(args_default, func.__defaults__))
                                      
        kwargs_positional_keys = [k for k in args_default if k not in kwargs]         # f(k1, k2=v2, ...) called as f(v1, v2, ...) 
        kwargs_positional = dict(zip(kwargs_positional_keys, args[num_static:]))
        num_kwargs_positional = len(kwargs_positional)
        
        kwargs = {**kwargs_positional, **kwargs}        
    else:
        num_kwargs_positional = 0
        kwargs_defaults = {}
        
    kwargs_defaults.update(kwonly_defaults)
                    
    full_kwargs = {**kwargs_defaults, **kwargs}
    static_attrs = {k: v for k, v in zip(code.co_varnames[:num_static], args)}    
    attrs = {**static_attrs, **full_kwargs} 
    
    pos_only_name = [p.name for p in inspect.signature(func).parameters.values() if p.kind.name == "VAR_POSITIONAL"]
    if pos_only_name:
        pos_only_name = f"*{pos_only_name[0]}"
    else:
        pos_only_name = "*pos_only"
    pos_only = {pos_only_name: len(args) - num_static - num_kwargs_positional}
    if add_pos_only_to_all:
        attrs.update(pos_only)
    
    key_opts = dict(kwargs_defaults = kwargs_defaults,
                    kwargs = kwargs,      
                    kwargs_full = full_kwargs,
                    args = static_attrs,
                    pos_only = pos_only,
                    all = attrs
    )
    return key_opts

def merge_wrapper_signatures(wrapper, wrapper_params):
    """
    Add parameters from the wrapper function (called by @functools.wraps) to its signature.
    Wrapper params are assumed to be keyword-only.
    
    Attrs:
            - wrapper:           wrapper function
            - wrapper_params:    params of wrapper to be added to the signature.
    """
    sig = inspect.signature(wrapper, follow_wrapped=True)
    sigw = inspect.signature(wrapper, follow_wrapped=False)
    params = tuple(sig.parameters.values())
    params_extra = tuple(p for p_name, p in sigw.parameters.items() if p_name in wrapper_params)
    if len(params) > 0 and params[-1].kind.name == "VAR_KEYWORD": # f(any_args, **kwargs)
        params_full = (*params[:-1], *params_extra, params[-1])
    else:
        params_full = (*params, *params_extra)
    return sig.replace(parameters=params_full)

def get_class_that_defined_method(meth, default=None):
    if isinstance(meth, functools.partial):
        return get_class_that_defined_method(meth.func)
    if inspect.ismethod(meth) or (inspect.isbuiltin(meth) and getattr(meth, '__self__', None) is not None and getattr(meth.__self__, '__class__', None)):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = getattr(meth, '__func__', meth)  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0],
                      None)
        if isinstance(cls, type):
            return cls
    return getattr(meth, '__objclass__', default)  # handle special descriptor objects

def delete_cls_self_arg(args):
    if args:
        if inspect.isclass(args[0]) or isinstance(args[0], get_class_that_defined_method(func, default=type(None))): # classmethod or instance method
            args = args[1:]
    return args


"""
Encode dictionaries in strings and decode them. Useful for automatically asigning filenames.
Includes hashing.
"""
import numpy as np
import math
import os
import hashlib

##############################################################################################################################
"""                                                  I. Getopt utils                                                       """
##############################################################################################################################

def encoder(x, ndigits=2, iterables=(list, tuple, np.ndarray), iterable_maxsize=3):
    """x -> string version of x"""
    if isinstance(x, float):
        if math.isinf(x) or math.isnan(x):
            return str(x)
        elif x == int(x):
            return str(int(x))
        else:
            return str(round(x, ndigits=ndigits)).replace('.', '--')
    elif isinstance(x, int):
        return str(x)
    elif callable(x):
        return x.__name__
    elif isinstance(x, dict):
        if len(x) >= iterable_maxsize:
            return "{}-values".format(str(len(x)))
        else:
            return dict_to_id(x, ndigits=ndigits, join_char="-", iterable_maxsize=iterable_maxsize)
    elif isinstance(x, iterables):
        if len(x) >= iterable_maxsize:
            return "{}-values".format(str(len(x)))
        else:
            return '-'.join([encoder(sub_x) for sub_x in x])
    else:
        return str(x)

def decoder(x, iterables=(list, tuple, np.ndarray)):
    """string version of x -> x"""
    if x.lower() == "none":
        return None
    elif x.lower() == "false":
        return False
    elif x.lower() == "true":
        return True
    elif "--" in x:
        return float(x.replace("--", "."))
    elif "." in x:
        try:
            return float(x)
        except:
            return x
    elif isinstance(x, iterables):
        return [decoder(sub_x) for sub_x in x]
    else:
        try:
            return int(x)
        except:
            return x

def getopt_printer(opts):
    """Prints getopt input in a readable way."""
    print('\n'.join(f'{opt} => {arg}' for opt, arg in (("Args", "Values"), *opts)))

def dict_to_id(*args, ndigits=2, join_char="_", iterable_maxsize=3, **kwargs):
    """Generate ID of the form k1-v1_k2-v2... for k_i, v_i keys and values of the dictionary d or the kwargs."""
    key_formatter = lambda k: k.replace("_", "-")
    d = args[0] if len(args) > 0 else kwargs
    return join_char.join([f"{key_formatter(k)}-{encoder(d[k], ndigits=ndigits, iterable_maxsize=iterable_maxsize)}" for k in sorted(d.keys())])

def id_to_dict(identifier):
    """Inverse of dict_to_id."""
    s = identifier.split("/")[-1] # retain filename only
    s = os.path.splitext(s)[0] # remove extension
    d = {}
    for split in s.split("_"):
        var_value = split.split("-")
        if len(var_value) > 1:
            if "" in var_value: # value is a float
                var_value_arr = np.array(var_value)
                idx_dot = np.argwhere(var_value_arr == "")[0, 0]
                key_idx = 0 if idx_dot == 2 else slice(0, idx_dot-2)
                d["-".join(var_value[key_idx])] = decoder(f"{var_value_arr[idx_dot-1]}--{var_value_arr[idx_dot+1]}")
            else:
                d["-".join(var_value[:-1])] = decoder(var_value[-1])
    return d

def id_updater(filename, update_dict, mode="add"):
    """
    Modifies filename by updating the underlying dict.
    Attrs:
        - filename:    id to be modified
        - update_dict: dict to use for updating the id. if update_dict={} => filename rearranged according to other_utils.dict_to_id.
        - mode:        - "add":    add update_dict to the id.
                       - "delete": delete update_dict from the id.
    Returns modified filename.
    """
    split_dirs = filename.split("/")
    parentDir = "/".join(split_dirs[:-1])
    file = split_dirs[-1]
    d = id_to_dict(file)
    if mode == "add":
        d.update(update_dict)
    elif mode == "delete":
        d = {k: v for k, v in d.items() if k not in update_dict.keys()}
    var_values = [part.split("-") for part in file.split("_")]
    init = "_".join([part[0] for part in var_values if len(part) == 1])
    ext = os.path.splitext(file)[1]
    new_filename = os.path.join(parentDir, f"{init}_{dict_to_id(d)}{ext}")

    return new_filename

def id_renamer(update_dict, parentDir, key=None, mode="add"):
    """
    Modifies id of files in parentDir by updating the underlying dict.
    Attrs:
        - update_dict: dict to use for updating the id
        - parentDir: folder where files are located.
        - key: string contained in the file for it to be modified.
        - mode:   - "add": add update_dict to the id.
                  - "delete": delete update_dict from the id.
    Returns #modified files.
    NOTE: If update_dict={} => filenames will be rearranged according to other_utils.dict_to_id.
    """
    r = 0
    for file in os.listdir(parentDir):
        if key is None or key in file:
            old_filename = os.path.join(parentDir, file)
            new_filename = id_updater(old_filename, update_dict, mode=mode)
            os.rename(old_filename, new_filename)
            r += 1
    return r


##############################################################################################################################
"""                                                     II. Other.                                                         """
##############################################################################################################################

def hash_file(filename, blocksize=65536):
    """
    Returns the SHA256 hash of the file.
    Autogenerated by copilot. Do not trust it.
    """
    hasher = hashlib.sha256()
    with open(filename, 'rb') as f:
        buf = f.read(blocksize)
        while len(buf) > 0:
            hasher.update(buf)
            buf = f.read(blocksize)
    return hasher.hexdigest()

def hash_string(s):
    """
    Returns the SHA256 hash of the string.
    """
    return hashlib.sha256(s.encode('utf-8')).hexdigest()



"""
Handle data (store/move/delete/load...)
"""

import numpy as np
import pandas as pd
import os
from copy import deepcopy
from pathlib import Path
import pickle
import bz2
import lzma
import json
import zipfile

parent_dir =  'data'


##############################################################################################################################
"""                                                       I. Save                                                          """
##############################################################################################################################

class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpEncoder, self).default(obj)

def save_preprocess(filename, saving_type, parent_dir):
    """Creates parent dir if does not exist and ensures filename has the saving_type extension."""
    _parent_dir, _filename = os.path.split(filename)
    if _parent_dir:  # directory specified in filename
        parent_dir = _parent_dir
        filename = _filename
    Path(parent_dir).mkdir(parents=True, exist_ok=True)
    filename = filename if filename.endswith('.{}'.format(saving_type)) else '{}.{}'.format(filename, saving_type)
    return parent_dir, filename
        
def create_saving_func(file_opener, file_writer, saving_type, **kwargs):
    def saving_func(data, filename, parent_dir=parent_dir):
        """Can specify directory in filename or in parent_dir."""
        parent_dir, filename = save_preprocess(filename, saving_type, parent_dir)
        with file_opener(os.path.join(parent_dir, filename), 'w') as f:
            file_writer(data, f, **kwargs)
        return
    return saving_func

save_json = create_saving_func(open, json.dump, 'json', cls=NpEncoder)
save_bz2 = create_saving_func(bz2.BZ2File, pickle.dump, 'pbz2')
save_lzma = create_saving_func(lzma.LZMAFile, pickle.dump, 'lzma')

def save_npz(arr, filename, parent_dir=parent_dir, key="arr"):
    np.savez_compressed(os.path.join(parent_dir, filename), **{key:arr})
    return

def save_csv(data, filename, parent_dir=parent_dir, **df_kwargs):
    """
    If data is a numpy array => if 2D => data
                                   1D => [data]
    (required by np.savetxt)
    """
    parent_dir, filename = save_preprocess(filename, "csv", parent_dir)
    if isinstance(data, (pd.core.frame.DataFrame, pd.core.series.Series)):
        data.to_csv(filename, **kwargs)
    else:
        np.savetxt(filename, data, delimiter=',', **kwargs)
    return


##############################################################################################################################
"""                                                       II. Load                                                         """
##############################################################################################################################

def create_loading_func(file_opener, file_loader, extra_processing=None, apply_defaults=None):
    """
    - extra_processing = List of process that could be applied to the file.
    - apply_defaults = dict:{str: bool}. The key is set as a function variable, the value indicates whether to apply the process named by the key.
    """
    if extra_processing is None:
        def loading_func(path):
            with file_opener(path, 'rb') as f:
                return file_loader(f)
    else:        
        def loading_func(path):
            args = {key:val for key,val in locals().items() if key not in  ('path', *create_loading_func.__code__.co_varnames)}
            with file_opener(path, 'rb') as f:
                data = file_loader(f)
            
            for condition, process in zip(args.values(), extra_processing):
                if condition:
                    data = process(data)
            return data
        
        # I know this sould not be done this way, but I wanted to check it can
        code_data = loading_func.__code__
        num_vars = code_data.co_argcount
        num_vars_new = len(apply_defaults)
        new_code = code_data.replace(co_varnames=(*code_data.co_varnames[:num_vars], *apply_defaults.keys(), *code_data.co_varnames[num_vars:]),
                                     co_argcount=num_vars + num_vars_new, 
                                     co_nlocals=code_data.co_nlocals + num_vars_new)
        loading_func.__code__ = new_code
        loading_func.__defaults__ = tuple(apply_defaults.values())
    return loading_func

def int_keys(dictionary):
     return {int(key):val for key,val in dictionary.items()}

load_json = create_loading_func(open, json.load, extra_processing=[int_keys], apply_defaults={'int_keys':True})
load_bz2 = create_loading_func(bz2.BZ2File, pickle.load)
load_lzma = create_loading_func(lzma.LZMAFile, pickle.load)

def load_npz(path, key=None):
    data = np.load(path)
    key = [*data.keys()][0] if key is None else key
    return data[key]

def load_csv(path, mode="pandas", **kwargs):
    if mode == "pandas":
        return pd.read_csv(path, **kwargs)
    elif mode == "numpy":
        return np.loadtxt(path, delimiter=",", **kwargs)
    else:
        raise ValueError(f"mode {mode} not valid. Available: 'panda', 'numpy'.")
        

##############################################################################################################################
"""                                                   III. Compression                                                     """
##############################################################################################################################

def compress_files(rootDir=parent_dir, extension='.json', compress_to='lzma', min_size=0, loading_key=None):
    """
    Searches in a directory and all its subdirectories files with a certain extension and compresses them.
    
    Attributes:
    - rootDir: The root directory.
    - extension: Extension setting which files should be compressed. Available options: '.json', '.npz'.
    - compress_to: File type after compression. Available options: 'lzma', 'bz2' (or pbz2).
    - min_size: Minimum size for applying compression (in MB).
    - loading_key: Key for retrieving the data in a npz file. If None, retrieves the data corresponding to the first key.
    
    Returns: Dict containing the name of the files that could not be processed ('bad compression') or those that were corrupted and had to be deleted('deleted files').
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell': # script being run in Jupyter notebook
            from tqdm.notebook import tqdm
        elif shell == 'TerminalInteractiveShell': #script being run in iPython terminal
            from tqdm import tqdm
    except NameError:
            from tqdm import tqdm # Probably runing on standard python terminal.
            
    # Get all the paths with the desired extension and minimum size.
    files = []

    for dirpath, subdirList, fileList in os.walk(rootDir):
        files += [os.path.join(dirpath, file) for file in fileList if os.path.splitext(file)[1] == extension]
    
    if min_size > 0:
        files = [file for file in files if os.path.getsize(file) > min_size*1e6]
    
    # Load files, compress, save and delete if compressed_file == pre_compressed_file
    if extension == '.json':
        loader = load_json
    elif extension == '.npz':
        loader = load_npz
    else:
        raise ValueError("extension = '{}' not valid. Available options: '.json', '.npz'".format(extension))
        
    if compress_to == 'lzma':
        compressor = save_lzma
        load_compressor = load_lzma
    elif compress_to == 'bz2':
        compressor = save_bz2
        load_compressor = load_bz2
    else:
        raise ValueError("compress_to = '{}' not valid. Available options: 'lzma', 'bz2'".format(extension))
    
    not_correctly_processed = {'bad_compression': []}
    pbar = tqdm(range(len(files)))
    
    if extension == '.json':
        for file in files:
            try:
                pre_compressed = loader(file, int_keys=True)
            except ValueError:
                pre_compressed = loader(file, int_keys=False)
            
            new_filename = '{}.{}'.format(os.path.splitext(file)[0], compress_to)
            compressor(pre_compressed, new_filename, parent_dir="")
            compressed_file = load_compressor(new_filename)
            
            if compressed_file == pre_compressed:
                os.remove(file)
            else:
                not_correctly_processed['bad_compression'].append(file)
            
            pbar.refresh()
            print(pbar.update())
            
    if extension == '.npz':
        not_correctly_processed['deleted_files'] = []
        for file in files:
            try:
                pre_compressed = loader(file, key=loading_key)
            except zipfile.BadZipFile: # corrupted file
                os.remove(file)
                not_correctly_processed['deleted_files'].append(file)
                continue
                
            new_filename = '{}.{}'.format(os.path.splitext(file)[0], compress_to)
            compressor(pre_compressed, new_filename, parent_dir="")
            compressed_file = load_compressor(new_filename)
            
            if np.all(compressed_file == pre_compressed):
                os.remove(file)
            else:
                not_correctly_processed['bad_compression'].append(file)
            pbar.refresh()
            print(pbar.update())
    
    return not_correctly_processed
        


from setuptools import setup

setup(
    name='tidypath',
    version='1.2.0',
    author="Jorge Medina Hernández",
    author_email='medinahdezjorge@gmail.com',
    packages=['tidypath'],
    url='https://github.com/medinajorge/tidypath',
    download_url='https://github.com/medinajorge/tidypath/archive/refs/tags/v1.0.5.tar.gz',
    description="Automatically store/load data in a tidy, efficient way.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords=['tidy', 'project organization', 'project', 'organization', 'path', 'storage'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Office/Business",
        "Intended Audience :: Science/Research",
    ],
    python_requires=">=3",
    install_requires=[
        'numpy',
        'pandas',
    ],
    extras_require={
        "matplotlib": "matplotlib",
        "plotly": ["plotly", "kaleido"],
    },
)



"""
Creates paths for automatically organizing files.
Allows filename modification: If arg added (deleted) to func => add (delete) arg to stored function output files.
Hash filename of long paths.

Two schemes:
    1. By module => Module -> submodules -> class tree (until classes in module by default) -> func
    2. By class  => Class organization is a class to be inherited. Sets path: parentDir -> subfolder (optionally specified) -> inheritance_tree -> func_name.
(1) is the most appropiate.
"""
import os
import sys
import inspect
import warnings
import numpy as np
import pandas as pd
import datetime
from copy import deepcopy
from collections.abc import Iterable
from pathlib import Path
from .fmt import dict_to_id, encoder, hash_string
from .inspection import get_class_that_defined_method

dataDir = "data"
figDir = "figs"

def time_since_last_update(parent_dir, ext='lzma'):
    """
    Returns a pandas Series with the time since last update of each file in parent_dir.
    """
    t = {}
    today = datetime.datetime.now()
    for f in os.listdir(parent_dir):
        if ext is None or f.endswith(ext):
            path = os.path.join(parent_dir, f)
            last_update = today - pd.to_datetime(Path(path).stat().st_mtime, unit='s')
            t[path] = last_update
    return pd.Series(t).sort_values()

def hash_path(path):
    """
    Hashes the filename (keeping the extension and parentDir) of a path.
    """
    parentDir = os.path.dirname(path)
    filename = os.path.basename(path)
    filename, ext = os.path.splitext(filename)
    filename = hash_string(filename)
    return os.path.join(parentDir, f"{filename}{ext}")

def module_path(func=None, depth=3, include_classes="file", skip=1):
    """
    returns module path and func_name.

    include_classes:   'file': path: module -> class tree contained in func file -> func_name.
                       'all': path: module -> full class tree -> func_name.
    """
    if func is None:
        frame = sys._getframe(depth)
        func_name = frame.f_code.co_name
        module = frame.f_globals["__name__"]
    else:
        func_name = func.__name__
        module = func.__module__
    if include_classes:
        func_class = get_class_that_defined_method(func)
        if func_class is not None:
            if include_classes == "file":
                clsmembers = inspect.getmembers(sys.modules[module], inspect.isclass)
                cls_names, clss = zip(*clsmembers)
                class_tree = "/".join([c.__name__.lower() for c in func_class.mro()[:-skip][::-1] if c in clss])
            elif include_classes == "all":
                "/".join([c.__name__.lower() for c in func.__class__.mro()[:-skip][::-1]])
            else:
                raise ValueError(f"include_classes: {include_classes} not valid. Available: 'file', 'all', ''.")
        else:
            class_tree = ""
    else:
        class_tree = ""
    path = os.path.join(*module.split(".")[1:], class_tree, func_name)
    return path

def saving_path(Dir, ext, keys={}, subfolder="", return_dir=False, funcname_in_filename=False, iterable_maxsize=3, **kwargs):
    """Tree path: subfolder -> module -> (classes) -> func_name."""
    parentDir = os.path.join(Dir, subfolder, module_path(**kwargs))
    Path(parentDir).mkdir(exist_ok=True, parents=True)
    if return_dir:
        return parentDir
    else:
        if funcname_in_filename:
            filename = parentDir.split("/")[-1] + f"_{dict_to_id(keys, iterable_maxsize=iterable_maxsize)}_.{ext}"
        else:
            filename = f"{dict_to_id(keys, iterable_maxsize=iterable_maxsize)}_.{ext}"
        return os.path.join(parentDir, filename)

def figpath(ext="png", **kwargs):
    return saving_path(figDir, ext, **kwargs)

def datapath(ext="lzma", **kwargs):
    return saving_path(dataDir, ext, **kwargs)

def filename_modifier(process_filename, func=None, directory=None, check_first=True, overwrite=False, **args):
    """
    Base function for adding/deleting/modifying function args encoded in the output filenames.

    Attrs:
        - process_filename:    function f(file, k_name, k, v) -> new_filename.
                               - k, v are the key and value of an arg.
                               - k_name is the string encoding of k. (replace '_' -> '-')
        - func (callable):     function (wrapped by savedata or savefig) for retrieving the corresponding directory where the outputs are stored.
        - directory (str):     directory where files are to be modified. It is recommended to pass the function instead of the directory.
        - check_first (bool):  whether to check the new filenames before applying any changes.
        - overwrite (bool):    whether to overwrite files when in conflict.
                               Examples:   add/modify/delete an arg -> two filenames are the same -> one remains.
        - args (kwargs):       Arguments to add/delete/modify in the filename.
    """
    if func is not None:
        if directory is not None:
            raise RuntimeError("Please provide 'func' or 'directory' but not both.")
        else:
            if func.__out__ == "data":
                directory = datapath(func=func, return_dir=True)
            else:
                directory = figpath(func=func, return_dir=True)
    elif directory is None:
        raise RuntimeError("Provide 'func' (preferred) or 'directory'.")

    args_sorted = {k: args[k] for k in sorted(args)}
    abort_changes = False
    avoid_files = [".ipynb_checkpoints", "__pycache__"]
    for k, v in args_sorted.items():
        check_first_k = deepcopy(check_first)
        if abort_changes:
            break
        else:
            k_name = encoder(k).replace("_", "-")
            for file in os.listdir(directory):
                if file not in avoid_files:
                    new_filename = process_filename(file, k_name, k, v)
                    if new_filename is not None:
                        new_path = os.path.join(directory, new_filename)
                        if Path(new_path).exists() and not overwrite:
                            raise RuntimeError(f"'{new_filename}' already existing before modifying '{file}'. To delete repeated files pass 'overwrite=True'.")
                        else:
                            if check_first_k:
                                update_file = 0
                                while update_file not in ["y", "yes", "n", "no"]:
                                    update_file = input("Filename change example:\n\n'{}' -> '{}'\n\nProceed? [y/n]".format(file, new_filename)).lower()
                                check_first_k = False
                            if update_file in ["y", "yes"]:
                                os.rename(os.path.join(directory, file),
                                          os.path.join(directory, new_filename))
                            else:
                                warnings.warn("Aborted filename changes.", RuntimeWarning)
                                abort_changes = True
                                break
    return

def add_arg(func=None, directory=None, check_first=True, overwrite=False, **args):
    """
    Encodes new function args into filename.
    Useful when:
        (1) Adding a new arg to a function wrapped by savedata. Allows loading the available data without recomputing.
        (2) Modifying figure names produced by a function.

    Attrs:
        - args (kwargs):       new arguments to encode into filename.
        - func (callable):     function (wrapped by savedata or savefig) for retrieving the corresponding directory where the outputs are stored.
        - directory (str):     directory where files are to be modified. It is recommended to pass the function instead of the directory.
        - check_first (bool):  whether to check the new filenames before applying any changes.
        - overwrite (bool):    whether to overwrite files when in conflict.
                               Example:   add an arg -> two filenames are the same -> one remains.
    """
    def process_filename(file, k_name, k, v):
        arg_val = dict_to_id({k:v})
        if arg_val not in file:
            args_in_file = np.sort(os.path.splitext(file)[0].split("_")[:-1] + [k])
            pos_new_arg = np.where(args_in_file == k)[0][0]
            if pos_new_arg == 0:
                new_filename = f"{arg_val}_{file}"
            else:
                if pos_new_arg + 1 == len(args_in_file):
                    insert_before = "."
                else:
                    insert_before = args_in_file[pos_new_arg + 1]
                new_filename = file.replace(f"_{insert_before}", f"_{arg_val}_{insert_before}")
            return new_filename
        else:
            return None
    filename_modifier(process_filename, func=func, directory=directory, check_first=check_first, overwrite=overwrite, **args)
    return

def delete_arg(func=None, arg=None, directory=None, check_first=True, overwrite=False):
    """
    Delete encoded function arg from filename.
    Useful when:
        (1) Deleting an arg from a function wrapped by savedata. Allows loading the available data without recomputing.
        (2) Modifying figure names produced by a function.

    Attrs:
        - arg (str or iterable):       key of an arg (k in k=v) to be deleted, or an iterable containing keys.
        - func (callable):             function (wrapped by savedata or savefig) for retrieving the corresponding directory where the outputs are stored.
        - directory (str):             directory where files are to be modified. It is recommended to pass the function instead of the directory.
        - check_first (bool):          whether to check the new filenames before applying any changes.
        - overwrite (bool):            whether to overwrite files when in conflict.
                                       Examples:   Delete an arg -> two filenames are the same -> one remains.
    """
    if isinstance(arg, str):
        arg_dict = {arg: None}
    elif isinstance(arg, Iterable):
        if all(isinstance(a, str) for a in arg):
            arg_dict = {a: None for a in arg}
        else:
            raise ValueError(f"arg {arg} not valid. Must be a string or an iterable of strings.")
    else:
        raise ValueError(f"arg {arg} not valid. Must be a string or an iterable of strings.")

    def process_filename(file, k_name, k, v):
        if file.startswith(f"{k_name}-"):
            return "_".join(file.split("_")[1:])
        else:
            mid_file_k_name = f"_{k_name}-"
            if mid_file_k_name in file:
                arg_val_encoded = "{}{}".format(mid_file_k_name, file.split(mid_file_k_name)[1].split("_")[0])
                return file.replace(arg_val_encoded, "")
            else:
                return None
    filename_modifier(process_filename, func=func, directory=directory, check_first=check_first, overwrite=overwrite, **arg_dict)
    return

def modify_arg(func=None, directory=None, check_first=True, overwrite=False, **args):
    """
    Modified encoded function arg from filename.
    Useful when:
        (1) Renaming an arg value with a more suitable one (more comprehensible, better defined...), leaving the func output unchanged. Allows loading the available data without recomputing.
        (2) Modifying figure names produced by a function.

    Attrs:
        - args (kwargs):       arguments to modify in the filename.
        - func (callable):     function (wrapped by savedata or savefig) for retrieving the corresponding directory where the outputs are stored.
        - directory (str):     directory where files are to be modified. It is recommended to pass the function instead of the directory.
        - check_first (bool):  whether to check the new filenames before applying any changes.
        - overwrite (bool):    whether to overwrite files when in conflict.
                               Examples:   Modify an arg -> two filenames are the same -> one remains.
    """
    def process_filename(file, k_name, k, v):
        if file.startswith(f"{k_name}-"):
            return "_".join([dict_to_id({k:v})] + file.split("_")[1:])
        else:
            mid_file_k_name = f"_{k_name}-"
            if mid_file_k_name in file:
                arg_val_encoded = "{}{}".format(mid_file_k_name, file.split(mid_file_k_name)[1].split("_")[0])
                return file.replace(arg_val_encoded, "_{}".format(dict_to_id({k:v})))
            else:
                return None

    filename_modifier(process_filename, func=func, directory=directory, check_first=check_first, overwrite=overwrite, **args)
    return

class Organizer():
    dataDir = dataDir
    figDir = figDir
    subfolder = ""

class ClassPath(Organizer):
    """
    Updates fig and data dir of the class as follows:
    parentDir -> subfolder -> inheritance_tree -> func_name
    """

    @classmethod
    def inheritance_path(cls, skip=2):
        return "/".join([c.__name__.lower() for c in cls.mro()[:-skip][::-1]])

    @classmethod
    def create_dir(cls, dirKey, depth=2):
        """Creates tree: parentDir -> subfolder -> inheritance_tree -> func_name"""
        path = os.path.join(getattr(cls, f"{dirKey}Dir"),
                            cls.subfolder,
                            cls.inheritance_path(),
                            sys._getframe(depth).f_code.co_name
                           )
        Path(path).mkdir(exist_ok=True, parents=True)
        return path

    @classmethod
    def create_figDir(cls):
        return cls.create_dir("fig")

    @classmethod
    def create_dataDir(cls):
        return cls.create_dir("data")
